/**
    (C) Copyright 2010 Dominik Wagenfuehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see
    <http://www.gnu.org/licenses/>.
*/

// Own
////////
#include "key.hh"
#include "trace.hh"

// Qt
////////
#include <QRegExp>

// Sys
////////
#include <iostream>

/// Constructor.
Key::Key( const QChar& id )
{
    set(id);
}

/// Comparison operator.
bool Key::operator==( const Key& key ) const
{
    return ( m_keyId == key.m_keyId );
}

/// Set new key.
bool Key::set( const QChar& id )
{
    bool ok = false;

    Trace::print( MTP_KEY, STP_KEY_SET,
                  "Key::set Start " + QString(id) );

    if ( QRegExp( "^[a-z]$" ).exactMatch( id ) )
    {
        m_keyId = id;
        ok = true;
    }
    else
    {
        std::cerr << "(EE) Key::set "
                  << " Character '"
                  << id.toAscii()
                  << "' is not a valid key character."
                  << std::endl;
    }

    Trace::print( MTP_KEY, STP_KEY_SET,
                  "Key::set End " + QString::number((int)ok) );
    
    return ok;
}
