(    
    {C} Copyright 2010 Harald Demel and Martin Keiblinger

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or {at your option} any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see
    <http://www.gnu.org/licenses/>.
)

create distMap width @ height @ * dup * cells allot
create bluePlayersMap width @ height @ * cells allot
create redPlayersMap width @ height @ * cells allot
create lastSeenMap width @ height @ * cells allot

: index ( n1 n2 -- n )
  ( n1 = x of current position )
  ( n2 = y of current position )
  ( n = index in list )
  width @ * +
  ;

: initLastSeenMap ( -- )
  height @ 0 do
    width @ 0 do
      10000 i j index cells lastSeenMap + !
    loop
  loop
  ;

initLastSeenMap

variable round
0 round !

 ( Words ) 

: distMap@ ( n1 n2 n3 n4 -- n )
  ( n1 = x of current position )
  ( n2 = y of current position )
  ( n3 = x of target position )
  ( n4 = y of target position )
  ( n = fetched value )
  index -rot
  index width @ height @ * * +
  cells distMap + @
  ;


: distMap! ( n1 n2 n3 n4 n5 -- )
  ( n1 = x of current position )
  ( n2 = y of current position )
  ( n3 = x of target position )
  ( n4 = y of target position )
  ( n5 = value to store )
  >r
  index -rot
  index width @ height @ * * +
  cells distMap + r> swap !
  ;

: getNeighbour ( n1 n2 n3 -- n4 n5 )
  ( n1 = x )
  ( n2 = y )
  ( n3 = the direction of neighbour as in libplayer/direction.hh )
  ( n4 = x of neighbour )
  ( n5 = y of neighbour )
   swap -rot ( y x n )
  dup 2dup 1 > swap 5 < and swap 5 > negate + ( y x n dx )
  rot + -rot ( new_x y n )
  dup 2dup 1 = rot 2 = rot 8 = or or swap dup 3 > swap 7 < and negate + ( new_x y dy )
  + ( new_x new_y )
  ;

: validField ( n1 n2 -- f )
  ( n1 = x )
  ( n2 = y )
  ( f  = is field valid )
  dup -1 > swap height @ < and swap dup -1 > swap width @ < and and
  ;

: emptyField ( n1 n2 -- f )
  ( n1 = x )
  ( n2 = y )
  ( f  = is field empty )
  getField 1 =
  ;

: validMove ( n1 n2 n3 n4 -- f )
  ( n1 = x from )
  ( n2 = y from )
  ( n3 = x from )
  ( n4 = y from )
  ( f  = is move valid )
  >r >r r@ -rot swap ( n3 n2 n1 )
  r> r> ( n3 n2 n1 n3 n4 )
  swap over ( n3 n2 n1 n4 n3 n4 )
  emptyField -rot
  emptyField and -rot
  emptyField and
  ;

: calcDistances ( n1 n2 -- )
  ( n1 = x of current position )
  ( n2 = y of current position )
  height @ 0 do
    width @ 0 do
      2dup i j 100 distMap!
    loop
  loop
  2dup 2dup 0 distMap!
  100 0 do
    i ( cur_x cor_y dist )
    height @ 0 do
      width @ 0 do
        >r 2dup r> -rot ( cur_x cor_y dist cur_x cor_y )
        i j distMap@ k = if
          9 1 do
            j k i getNeighbour ( cur_x cor_y dist x y )
            2dup validField if
              2dup j k 2swap validMove if
                rot >r 2over 2over 2over 2over distMap@ r@ 1+ min distMap! r> -rot
              then
            then
            2drop
          loop
        then
      loop
    loop
    drop
  loop
  2drop
  ;

: distMapCalc@ ( n1 n2 n3 n4 -- n )
  ( n1 = x of current position )
  ( n2 = y of current position )
  ( n3 = x of target position )
  ( n4 = y of target position )
  ( n = fetched value )
  2over 2over distMap@ 0= if
    2over 2over rot <> -rot <> or if
      2over calcDistances
    then
  then
  distMap@
  ;

: calcAllDistances ( -- )
  height @ 0 do
    width @ 0 do
      ." _" i . j . i j calcDistances
    loop
  loop
  ;

: calcBluePlayersMap ( -- )
  height @ 0 do
    width @ 0 do
      0 i j index cells bluePlayersMap + !
    loop
  loop

  0 getPlayersOfTeam
  begin
    while ( x y infoAge )
      >r over
      dup r@ + 1 + width @ min swap r@ - 0 max ( x y x+infoAge x-infoAge )
      2swap ( x+infoAge x-infoAge x y )
      r@ r@ 2swap ( x+infoAge x-infoAge infoAge infoAge x y )
      dup dup r@ + 1 + height @ min swap r> - 0 max ( x+infoAge x-infoAge infoAge infoAge x y y+infoAge y-infoAge )
      do ( x+infoAge x-infoAge infoAge infoAge x y )
        >r >r 2over ( x+infoAge x-infoAge infoAge infoAge x+infoAge x-infoAge )
        r> r> 2swap ( x+infoAge x-infoAge infoAge infoAge x y x+infoAge x-infoAge )
        do ( x+infoAge x-infoAge infoAge infoAge x y )
          i j emptyField if
            2over 2over
            i j distMapCalc@ swap over 1 - > if ( x+infoAge x-infoAge infoAge dist )
              dup * dup * + 1 +
              i j index cells bluePlayersMap + @
              dup 0= if
                drop 
              else 
                min
              then
              i j index cells bluePlayersMap + !
            else
              2drop
            then
          then
        loop
      loop
      2drop
      2drop 2drop
  repeat
  ;

: calcRedPlayersMap ( -- )
  height @ 0 do
    width @ 0 do
      0 i j index cells redPlayersMap + !
    loop
  loop

  1 getPlayersOfTeam
  begin
    while ( x y infoAge )
      >r over
      dup r@ + 1 + width @ min swap r@ - 0 max ( x y x+infoAge x-infoAge )
      2swap ( x+infoAge x-infoAge x y )
      r@ r@ 2swap ( x+infoAge x-infoAge infoAge infoAge x y )
      dup dup r@ + 1 + height @ min swap r> - 0 max ( x+infoAge x-infoAge infoAge infoAge x y y+infoAge y-infoAge )
      do ( x+infoAge x-infoAge infoAge infoAge x y )
        >r >r 2over ( x+infoAge x-infoAge infoAge infoAge x+infoAge x-infoAge )
        r> r> 2swap ( x+infoAge x-infoAge infoAge infoAge x y x+infoAge x-infoAge )
        do ( x+infoAge x-infoAge infoAge infoAge x y )
          i j emptyField if
            2over 2over
            i j distMapCalc@ swap over 1 - > if ( x+infoAge x-infoAge infoAge dist )
              dup * dup * + 1 +
              i j index cells redPlayersMap + @
              dup 0= if
                drop 
              else 
                min
              then
              i j index cells redPlayersMap + !
            else
              2drop
            then
          then
        loop
      loop
      2drop
      2drop 2drop
  repeat
  ;

: updateLastSeenMap ( -- )
  height @ 0 do
    width @ 0 do
      i j index cells lastSeenMap + ( addrCellOfLastSeenMap )
      i j canSeeField if
        0 swap !
      else
        dup @ 1 + swap !
      then
    loop
  loop
  ;

 ( DUMPS )

: dumpDistances ( n1 n2 -- )
  ( n1 = x of current position )
  ( n2 = y of current position )
  height @ 0 do
    width @ 0 do
      2dup i j distMap@ dup 100 < if ." _" then dup 10 < if ." _" then dup 100 = if ." ### " drop else . then
    loop
   ." | " cr
  loop cr
  2drop
  ;

: dumpAllDistances ( -- )
  height @ 0 do
    width @ 0 do
      i j dumpDistances
    loop
  loop
  ;

: dumpBlue ( -- )
  height @ 0 do
    width @ 0 do
      i j index cells bluePlayersMap + @ dup 100000 < if ." _" then dup 10000 < if ." _" then dup 1000 < if ." _" then dup 100 < if ." _" then dup 10 < if ." _" then .
    loop
   ." | " cr
  loop cr
  ;

: dumpRed ( -- )
  height @ 0 do
    width @ 0 do
      i j index cells redPlayersMap + @ dup 100000 < if ." _" then dup 10000 < if ." _" then dup 1000 < if ." _" then dup 100 < if ." _" then dup 10 < if ." _" then .
    loop
   ." | " cr
  loop cr
  ;

: dumpLastSeenMap ( -- )
  height @ 0 do
    width @ 0 do
      i j index cells lastSeenMap + @ dup 100000000 < if ." _" then dup 10000000 < if ." _" then dup 1000000 < if ." _" then dup 100000 < if ." _" then dup 10000 < if ." _" then dup 1000 < if ." _" then dup 100 < if ." _" then dup 10 < if ." _" then .
    loop
   ." | " cr
  loop cr
  ;

: dumpPlayerTeam ( -- )
  height @ 0 do
    width @ 0 do
      i j getPlayerTeam dup -1 > if ." 0" then .
    loop
   ." | " cr
  loop cr
  ;

 ( END DUPMS )

: calcLocalScore ( n1 n2 -- r )
  ( n1 = x )
  ( n2 = y )
  ( r  = score )
  getOwnTeam 0 = if
    2dup getToxic
    2dup index cells redPlayersMap + @
    dup 0 = if
      drop 0e
    else
      dup 2over index cells lastSeenMap + @ dup * dup * 2 + < if
        3 * 1+ d>f
        fdup f* fdup f* 1000e fswap f/
      else
        drop 0e
      then
    then
    2drop f*
  else
    2dup index cells bluePlayersMap + @
    dup 0 = if
      drop 0e
    else
      dup 2over index cells lastSeenMap + @ dup * dup * 2 + 
        ( 2dup swap . . ( DEBUG )
        < if
        3 * 1+ d>f
        fdup f* fdup f* 1000e fswap f/
      else
        drop 0e
      then
    then
    2drop
  then
  ;

: calcSimpleScore ( n1 n2 -- r )
  ( n1 = x )
  ( n2 = y )
  ( r  = score )
  0e ( - float: score )
  dup 101 + height @ min over 100 - 0 max do
    over dup 101 + width @ min swap 100 - 0 max do
      ( ." ( " i . j . ( DEBUG )
      i j calcLocalScore
      ( fdup f. ( DEBUG )
      2dup i j distMapCalc@ 1+ 2 * dup * d>f
      ( fdup f. ." / " ( DEBUG )
      f/ f+
    loop
  loop
  2drop
  ;

 ( C-called )

: initCalc ( -- )
  round @ 1 + round !

  ( ." team " getOwnTeam . cr
  0 getPlayersOfTeam ." blue " cr
  begin
    while
      swap rot . . . cr
  repeat

  1 getPlayersOfTeam ." red " cr
  begin
    while
      swap rot . . . cr
  repeat )

  calcBluePlayersMap
  ( dumpBlue )
  calcRedPlayersMap
  ( dumpRed )
  updateLastSeenMap
  ( dumpLastSeenMap )
  ;

: calcMove ( -- )
  ( ." round " round @ . cr   ( DEBUG )
  0 -100000000000000000000000000000000000000000000000000000000000e ( bestdir - float: bestscore=-inf )
  getOwnPos ( bestdir own_x own_y )
  ( ."  pos: " 2dup swap . . cr   ( DEBUG )
  9 0 do
    2dup i getNeighbour ( bestdir own_x own_y x y )
    2dup validField if
      2over 2over validMove if
        ( ."  " i . 2dup swap . .   ( DEBUG )
        calcSimpleScore ( bestdir own_x own_y - float:  score )
        ( ."  " fdup 1000e f* f. cr   ( DEBUG )
        fover fover f< if
          fswap fdrop
          rot drop i -rot
        else
          fdrop
        then
      else
        2drop
      then
    else
      2drop
    then
  loop
  2drop
  ( ."  choice: " dup . cr   ( DEBUG )
  setMove
  ;

: calcView ( -- )
  getLastView 2 + 8 mod 1+ setView
  ;
