#!/usr/bin/python
# -*- coding: iso-8859-1 -*-

"""
Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

"""

import socket
import field

from globals import *
from players import *

class communication:
    def __init__(self):
        self.adress = "localhost"
        self.port = 15000

    def receiveData(self):
        data = self.sock.recv(4048)
        #print "RECEIVED: " + data

        return data

    def receiveKeyData(self, key):
        data = self.receiveData()

        if data.find(key) != -1:
           result = self.findKeyData(data,key)
           #print "KEYED " + data + " <--> " +  data.split()[1][:-1] + " <--> " + result
           return result  
        else: 
           return self.receiveKeyData(key)

    def findKeyData(self, data,key):
        allMessages = data.split("|")

        for message in allMessages:
            if len(message.split()) > 0 and message.split()[0] == key:
                return message.split()[1] 
        
    def sendOkay(self):
        self.sock.sendall("ID_OKAY|")
    
    def sendMove(self, dir, view):
        self.sock.sendall("MOVE %s %s|" % (dir,view))

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(60)

        self.sock.connect((self.adress,self.port))

        recvID = int(self.receiveKeyData("ID"))
        self.sendOkay()

        map = None
        numPlayer = 1
        while map == None:
            data = self.receiveData()

            keyData = self.findKeyData(data,"CONNECTED")

            if keyData != None:
                numPlayer += 1
            else:
                map = self.findKeyData(data,"MAP")

        #print "CONNECT RESULT: %d %s" % (recvID ,map)

        return (recvID,map,numPlayer)

    def endGame(self):
        self.sock.close()
    
    def parseMessage(self,playerdata):
        data = self.receiveData()

        allMessages = data.split("|")

        for message in allMessages:
            #print "Parse message: " + message
            splitMessage = message.split()

            if len(splitMessage) > 0:
                parser = messageParser(playerdata)
                parser.parse(splitMessage)
               

           
class messageParser:
    def __init__(self,playerdata):
        self.playerdata = playerdata
        self.parser = {"DISCONNECT"   : self.DISCONNECT,
                       "START"        : self.START,
                       "DISCONNECTED" : self.DISCONNECTED,
                       "SEE_PLAYER"   : self.SEE_PLAYER,
                       "SET"          : self.SET,
                       "TOXIC"        : self.TOXIC,
                       "LIFE"         : self.LIFE,
                       "TEAMCHANGE"   : self.TEAMCHANGE,
                       "PING"         : self.PING
                      }

    def PING(self,messageList):
        pass

    def DISCONNECT(self,messageList):
        raise EndGame
    
    def DISCONNECTED(self,messageList):
        for player in self.playerdata["players"]:
            if player.id == int(messageList[0]):
                player.active = False
                break
        else:
            aktplayer = enemyplayer()
            aktplayer.id = int(messageList[0])
            aktplayer.active = False
            self.playerdata["players"].append(aktplayer)

    def TEAMCHANGE(self,messageList):
        for player in self.playerdata["players"]:
            if player.id == int(messageList[1]):
                player.team = messageList[0]
                if player.id == self.playerdata["me"].id:
                    self.playerdata["me"].team = messageList[0]
                break
        else:
            aktplayer = enemyplayer()
            aktplayer.id = int(messageList[1])
            aktplayer.team = messageList[0]
            self.playerdata["players"].append(aktplayer)
           
    def LIFE(self,messageList):
        for player in self.playerdata["players"]:
            if player.id == int(messageList[1]):
                player.life = float(messageList[0])
                if player.id == self.playerdata["me"].id:
                    self.playerdata["me"].life = float(messageList[0])
                break
        else:
            aktplayer = enemyplayer()
            aktplayer.id = int(messageList[1])
            aktplayer.life = float(messageList[0])
            self.playerdata["players"].append(aktplayer)

    def TOXIC(self,messageList):
        x = int(messageList[0].split(',')[0])
        y = int(messageList[0].split(',')[1])
        pos = position()
        pos.setPosXY(x,y)
        self.playerdata["map"].setToxic(pos,float(messageList[1]))


    def SET(self,messageList):
        self.playerdata["me"].pos.setPos(messageList[0])
        self.playerdata["me"].view = messageList[1]
        
        for player in self.playerdata["players"]:
            if player.id == self.playerdata["me"].id:
                player.pos.setPos(messageList[0])
                player.view = messageList[1]

    def SEE_PLAYER(self,messageList):
        for player in self.playerdata["players"]:
            if player.id == int(messageList[2]):
                player.pos.setPos(messageList[0])
                player.view = messageList[1]
                player.seenRound = self.playerdata["round"]
                break
        else:
            aktplayer = enemyplayer()
            aktplayer.id = int(messageList[2])
            aktplayer.pos.setPos(messageList[0])
            aktplayer.view = messageList[1]
            self.playerdata["players"].append(aktplayer)

    def START(self,messageList):
        self.playerdata["round"] = int(messageList[0])

    def parse(self,messageList):
        try:
            self.parser[messageList[0]](messageList[1:])
        except KeyError as e:
            print "%s not implementet" % messageList[0] 
