#!/usr/bin/python
# -*- coding: iso-8859-1 -*-
"""
Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

"""
import os
import networkx as nx
import globals
from globals import directions
import time

class field:
    def __init__(self,path):
        self.path = path

    def readField(self):
        filelines = open(self.path).readlines()

        self.field = []

        for line in filelines:
            self.field.append([])
            for field in list(line[:-1]):
                self.field[-1].append({"val" : field, "toxic" : 0 , "mag" : 0})
                if field == "#":
                    self.field[-1][-1]["mag"] = 0.1

        self.buildGraph()

    def getXSize(self):
        return len(self.field)
       
    def getYSize(self):
        return len(self.field[0])

    def buildGraph(self):
        self.gr = nx.Graph()

        for lIndex,line in enumerate(self.field):
            for sIndex,spot in enumerate(line):
                if self.NoWall(lIndex,sIndex) == True:
                    self.gr.add_node(self.getNodeName(lIndex,sIndex),x = lIndex, y = sIndex)
                    #self.gr.add_node_attribute(self.getNodeName(lIndex,sIndex),('position',(lIndex,sIndex)))


        for lIndex,line in enumerate(self.field):
            for sIndex,spot in enumerate(line):
                if lIndex != 0 and lIndex != len(self.field) -1 and sIndex != 0 and sIndex != len(line) - 1 and self.NoWall(lIndex,sIndex) == True: 
                    #print lIndex,sIndex
                    #North
                    self.addMainEdge(lIndex,sIndex,lIndex -1 ,sIndex)
                    #North East 
                    self.addDiagonalEdge(lIndex,sIndex,lIndex - 1, sIndex + 1)
                    #East
                    self.addMainEdge(lIndex,sIndex,lIndex,sIndex + 1)
                    #South East
                    self.addDiagonalEdge(lIndex,sIndex,lIndex + 1, sIndex + 1)
                    #South
                    self.addMainEdge(lIndex,sIndex,lIndex + 1,sIndex)
                    #Sout West
                    self.addDiagonalEdge(lIndex,sIndex,lIndex + 1, sIndex - 1)
                    #West
                    self.addMainEdge(lIndex,sIndex,lIndex,sIndex - 1)
                    #North West
                    self.addDiagonalEdge(lIndex,sIndex,lIndex - 1, sIndex - 1)

    def addMainEdge(self,lIndex,sIndex,lTargetIndex,sTargetIndex):
        if self.NoWall(lTargetIndex,sTargetIndex) == True:
                #print "Adding edge from (%d,%d) is %s to (%d,%d) is %s" % (lIndex,sIndex,self.field[lIndex][sIndex]["val"] ,lTargetIndex,sTargetIndex,self.field[lTargetIndex][sTargetIndex]["val"] )
                #print "Adding edge from (%s) to (%s)" % (self.getNodeName(lIndex,sIndex),self.getNodeName(lTargetIndex,sTargetIndex))
                
                self.gr.add_edge(self.getNodeName(lIndex,sIndex),self.getNodeName(lTargetIndex,sTargetIndex))
        
    def addDiagonalEdge(self,lIndex,sIndex,lTargetIndex,sTargetIndex):
        if self.NoWall(lTargetIndex,sTargetIndex) == True and self.field[lIndex][sTargetIndex]["val"] != "#" and self.field[lTargetIndex][sIndex]["val"] != "#" :
                self.gr.add_edge(self.getNodeName(lIndex,sIndex),self.getNodeName(lTargetIndex,sTargetIndex))

    def getSizeGraph(self):
        return len(self.gr.nodes())

    def getPosFromGraph(self,index):
        if index < len(self.gr.nodes()):
            pos = globals.position()
            pos.setPosXY(int(self.gr.nodes()[index].split("_")[0]),int(self.gr.nodes()[index].split("_")[1]))
            return pos

    def NoWall(self,lIndex,sIndex):
        return self.field[lIndex][sIndex]["val"] != "#"

    def getNodeName(self,lIndex,sIndex):
        return "%03d_%03d" % (lIndex,sIndex)

    def setToxic(self,pos,value):

        self.field[pos.y][pos.x]["toxic"] = value

    def setPlayer(self,pos):

        for line in self.field:
            for field in line:
                if field["val"] == "P":
                    field["val"] = " "

        self.field[pos.y][pos.x]["val"] = "P"
    
    def setBlue(self,pos):

        for line in self.field:
            for field in line:
                if field["val"] == "p":
                    field["val"] = " "

        self.field[pos.y][pos.x]["val"] = "p"
    
    def setRed(self,pos):

        for line in self.field:
            for field in line:
                if field["val"] == "R":
                    field["val"] = " "

        self.field[pos.y][pos.x]["val"] = "R"


    def __str__(self):
        str = ""
        for index,line in enumerate(self.field):
            str += " %02d " % index
            for field in line:
                str += field["val"]
            str += "\n"
        return str
    
    def moveFromTo(self,start,target):

        result = nx.shortest_path(self.gr,self.getNodeName(start.x,start.y),self.getNodeName(target.x,target.y))
        nexttarget = globals.position()
        nexttarget.x = int(result[1].split("_")[0])
        nexttarget.y = int(result[1].split("_")[1])
        return self.getDirection(start,nexttarget)

    def moveAway(self,pos,away):
        vek = (pos.x - away.x,pos.y - away.y)

        if vek[0] < 0:
            if vek[1] < 0:
                return directions[7]    #NORTH_WEST 
            elif vek[1] == 0:
                return directions[0]    #NORTH
            else:
                return directions[1]    #NORTH_EAST
        elif vek[0] == 0:
            if vek[1] < 0:
                return directions[6]    #WEST 
            elif vek[1] == 0:
                return directions[-1]    #NONE
            else:
                return directions[2]    #EAST
        else:
            if vek[1] < 0:
                return directions[3]    #SOUTH_WEST
            elif vek[1] == 0:
                return directions[4]    #SOUTH
            else:
                return directions[3]    #SOUTH_EAST 


    def getDirection(self,start,target):
        if start.x < target.x:
           if start.y < target.y:
              return directions[3]    #SOUTH_EAST
           elif start.y == target.y:
              return directions[4]    #SOUTH
           else:
              return directions[5]    #SOUTH_WEST
        elif start.x == target.x:
           if start.y < target.y:
              return directions[2]    #EAST
           elif start.y == target.y:
              return directions[-1]   #NONE
           else:
              return directions[6]    #WEST
        else:
           if start.y < target.y:
              return directions[1]    #NORTH_EAST
           elif start.y == target.y:
              return directions[0]    #NORTH
           else:
              return directions[7]    #NORTH_WEST 

def testMove(field,x1,y1,x2,y2):
    start  = globals.position()
    target = globals.position()
    
    start.setPosXY(x1,y1)
    target.setPosXY(x2,y2)

    print "%s -> %s: %s" %(str(start),str(target),field.moveFromTo(start,target))
    
if __name__ == "__main__":
    test = field("../freiesmagazin-contest3/maps/board1.map")
    #test = field("../freiesmagazin-contest3/maps/smallboard.map")
    test.readField()

    #print test.gr.edges()
    print len(test.gr.nodes())
    print len(test.gr.edges())

    testMove(test,2,2,1,1)
    testMove(test,2,2,1,2)
    testMove(test,2,2,1,3)
    testMove(test,2,2,2,3)
    testMove(test,2,2,3,3)
    testMove(test,2,2,3,2)
    testMove(test,2,2,3,1)
    testMove(test,2,2,2,1)

    print test
